@@ -4,53 +4,52 @@
 Defines a request dispatcher, a HTTP request handler, a HTTP server and a
 CGI request handler.
 
-Licensed under the Apache License, Version 2.0 (the "License");
-you may not use this file except in compliance with the License.
-You may obtain a copy of the License at
-
-   http://www.apache.org/licenses/LICENSE-2.0
-
-Unless required by applicable law or agreed to in writing, software
-distributed under the License is distributed on an "AS IS" BASIS,
-WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-See the License for the specific language governing permissions and
-limitations under the License.
-
+:authors: Josh Marshall, Thomas Calmant
+:copyright: Copyright 2017, Thomas Calmant
 :license: Apache License 2.0
-:version: 0.2.3
+:version: 0.3.1
+
+..
+
+    Copyright 2017 Thomas Calmant
+
+    Licensed under the Apache License, Version 2.0 (the "License");
+    you may not use this file except in compliance with the License.
+    You may obtain a copy of the License at
+
+        http://www.apache.org/licenses/LICENSE-2.0
+
+    Unless required by applicable law or agreed to in writing, software
+    distributed under the License is distributed on an "AS IS" BASIS,
+    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+    See the License for the specific language governing permissions and
+    limitations under the License.
 """
 
-# Module version
-__version_info__ = (0, 2, 3)
-__version__ = ".".join(str(x) for x in __version_info__)
-
-# Documentation strings format
-__docformat__ = "restructuredtext en"
-
-# ------------------------------------------------------------------------------
-# Local modules
-from jsonrpclib import Fault
-import jsonrpclib.config
-import jsonrpclib.utils as utils
-
 # Standard library
 import logging
 import socket
 import sys
 import traceback
 
-# Prepare the logger
-_logger = logging.getLogger(__name__)
-
 try:
     # Python 3
     # pylint: disable=F0401,E0611
     import xmlrpc.server as xmlrpcserver
+    # Make sure the module is complete.
+    # The "future" package under python2.7 provides an incomplete
+    # variant of this package.
+    SimpleXMLRPCDispatcher = xmlrpcserver.SimpleXMLRPCDispatcher
+    SimpleXMLRPCRequestHandler = xmlrpcserver.SimpleXMLRPCRequestHandler
+    resolve_dotted_attribute = xmlrpcserver.resolve_dotted_attribute
     import socketserver
 except (ImportError, AttributeError):
     # Python 2 or IronPython
     # pylint: disable=F0401,E0611
     import SimpleXMLRPCServer as xmlrpcserver
+    SimpleXMLRPCDispatcher = xmlrpcserver.SimpleXMLRPCDispatcher
+    SimpleXMLRPCRequestHandler = xmlrpcserver.SimpleXMLRPCRequestHandler
+    resolve_dotted_attribute = xmlrpcserver.resolve_dotted_attribute
     import SocketServer as socketserver
 
 try:
@@ -61,6 +60,24 @@ except ImportError:
     # pylint: disable=C0103
     fcntl = None
 
+# Local modules
+from jsonrpclib import Fault
+import jsonrpclib.config
+import jsonrpclib.utils as utils
+import jsonrpclib.threadpool
+
+# ------------------------------------------------------------------------------
+
+# Module version
+__version_info__ = (0, 3, 1)
+__version__ = ".".join(str(x) for x in __version_info__)
+
+# Documentation strings format
+__docformat__ = "restructuredtext en"
+
+# Prepare the logger
+_logger = logging.getLogger(__name__)
+
 # ------------------------------------------------------------------------------
 
 
@@ -114,7 +131,7 @@ def validate_request(request, json_config):
     params = request.get('params')
     param_types = (utils.ListType, utils.DictType, utils.TupleType)
 
-    if not method or not isinstance(method, utils.string_types) or \
+    if not method or not isinstance(method, utils.STRING_TYPES) or \
             not isinstance(params, param_types):
         # Invalid type of method name or parameters
         fault = Fault(-32600, 'Invalid request parameters or method.',
@@ -135,7 +152,7 @@ class NoMulticallResult(Exception):
     pass
 
 
-class SimpleJSONRPCDispatcher(xmlrpcserver.SimpleXMLRPCDispatcher, object):
+class SimpleJSONRPCDispatcher(SimpleXMLRPCDispatcher, object):
     """
     Mix-in class that dispatches JSON-RPC requests.
 
@@ -148,10 +165,19 @@ class SimpleJSONRPCDispatcher(xmlrpcserver.SimpleXMLRPCDispatcher, object):
         Sets up the dispatcher with the given encoding.
         None values are allowed.
         """
-        xmlrpcserver.SimpleXMLRPCDispatcher.__init__(
+        SimpleXMLRPCDispatcher.__init__(
             self, allow_none=True, encoding=encoding or "UTF-8")
         self.json_config = config
 
+        # Notification thread pool
+        self.__notification_pool = None
+
+    def set_notification_pool(self, thread_pool):
+        """
+        Sets the thread pool to use to handle notifications
+        """
+        self.__notification_pool = thread_pool
+
     def _unmarshaled_dispatch(self, request, dispatch_method=None):
         """
         Loads the request dictionary (unmarshaled), calls the method(s)
@@ -181,8 +207,8 @@ class SimpleJSONRPCDispatcher(xmlrpcserver.SimpleXMLRPCDispatcher, object):
                     continue
 
                 # Call the method
-                resp_entry = self._marshaled_single_dispatch(req_entry,
-                                                             dispatch_method)
+                resp_entry = self._marshaled_single_dispatch(
+                    req_entry, dispatch_method)
 
                 # Store its result
                 if isinstance(resp_entry, Fault):
@@ -205,8 +231,8 @@ class SimpleJSONRPCDispatcher(xmlrpcserver.SimpleXMLRPCDispatcher, object):
                 return result.dump()
 
             # Call the method
-            response = self._marshaled_single_dispatch(request,
-                                                       dispatch_method)
+            response = self._marshaled_single_dispatch(
+                request, dispatch_method)
             if isinstance(response, Fault):
                 # pylint: disable=E1103
                 return response.dump()
@@ -269,24 +295,38 @@ class SimpleJSONRPCDispatcher(xmlrpcserver.SimpleXMLRPCDispatcher, object):
             # Keep server configuration as is
             config = self.json_config
 
-        try:
-            # TODO - Use the multiprocessing if it is a notification
-            # Call the method
+        # Test if this is a notification request
+        is_notification = 'id' not in request or request['id'] in (None, '')
+        if is_notification and self.__notification_pool is not None:
+            # Use the thread pool for notifications
             if dispatch_method is not None:
-                response = dispatch_method(method, params)
+                self.__notification_pool.enqueue(
+                    dispatch_method, method, params)
             else:
-                response = self._dispatch(method, params, config)
-        except Exception as ex:
-            # Return a fault
-            fault = Fault(-32603, '{0}:{1}'.format(type(ex).__name__, ex),
-                          config=config)
-            _logger.error("Error calling method %s: %s", method, fault)
-            return fault.dump()
+                self.__notification_pool.enqueue(
+                    self._dispatch, method, params, config)
 
-        if 'id' not in request or request['id'] in (None, ''):
-            # It's a notification, no result needed
-            # Do not use 'not id' as it might be the integer 0
+            # Return immediately
             return None
+        else:
+            # Synchronous call
+            try:
+                # Call the method
+                if dispatch_method is not None:
+                    response = dispatch_method(method, params)
+                else:
+                    response = self._dispatch(method, params, config)
+            except Exception as ex:
+                # Return a fault
+                fault = Fault(-32603, '{0}:{1}'.format(type(ex).__name__, ex),
+                              config=config)
+                _logger.error("Error calling method %s: %s", method, fault)
+                return fault.dump()
+
+            if is_notification:
+                # It's a notification, no result needed
+                # Do not use 'not id' as it might be the integer 0
+                return None
 
         # Prepare a JSON-RPC dictionary
         try:
@@ -323,7 +363,7 @@ class SimpleJSONRPCDispatcher(xmlrpcserver.SimpleXMLRPCDispatcher, object):
                 except AttributeError:
                     # Resolve the method name in the instance
                     try:
-                        func = xmlrpcserver.resolve_dotted_attribute(
+                        func = resolve_dotted_attribute(
                             self.instance, method, True)
                     except AttributeError:
                         # Unknown method
@@ -344,8 +384,8 @@ class SimpleJSONRPCDispatcher(xmlrpcserver.SimpleXMLRPCDispatcher, object):
                 return fault
             except:
                 # Method exception
-                err_lines = traceback.format_exc().splitlines()
-                trace_string = '{0} | {1}'.format(err_lines[-3], err_lines[-1])
+                err_lines = traceback.format_exception(*sys.exc_info())
+                trace_string = '{0} | {1}'.format(err_lines[-2].splitlines()[0].strip(), err_lines[-1])
                 fault = Fault(-32603, 'Server error: {0}'.format(trace_string),
                               config=config)
                 _logger.exception("Server-side exception: %s", fault)
@@ -360,7 +400,7 @@ class SimpleJSONRPCDispatcher(xmlrpcserver.SimpleXMLRPCDispatcher, object):
 # ------------------------------------------------------------------------------
 
 
-class SimpleJSONRPCRequestHandler(xmlrpcserver.SimpleXMLRPCRequestHandler):
+class SimpleJSONRPCRequestHandler(SimpleXMLRPCRequestHandler):
     """
     HTTP request handler.
 
@@ -385,21 +425,34 @@ class SimpleJSONRPCRequestHandler(xmlrpcserver.SimpleXMLRPCRequestHandler):
             chunks = []
             while size_remaining:
                 chunk_size = min(size_remaining, max_chunk_size)
-                chunks.append(utils.from_bytes(self.rfile.read(chunk_size)))
+                raw_chunk = self.rfile.read(chunk_size)
+                if not raw_chunk:
+                    break
+                chunks.append(utils.from_bytes(raw_chunk))
                 size_remaining -= len(chunks[-1])
             data = ''.join(chunks)
 
+            try:
+                # Decode content
+                data = self.decode_request_content(data)
+                if data is None:
+                    # Unknown encoding, response has been sent
+                    return
+            except AttributeError:
+                # Available since Python 2.7
+                pass
+
             # Execute the method
-            response = self.server._marshaled_dispatch(data)
+            response = self.server._marshaled_dispatch(
+                data, getattr(self, '_dispatch', None), self.path)
 
             # No exception: send a 200 OK
             self.send_response(200)
-
         except:
             # Exception: send 500 Server Error
             self.send_response(500)
-            err_lines = traceback.format_exc().splitlines()
-            trace_string = '{0} | {1}'.format(err_lines[-3], err_lines[-1])
+            err_lines = traceback.format_exception(*sys.exc_info())
+            trace_string = '{0} | {1}'.format(err_lines[-2].splitlines()[0].strip(), err_lines[-1])
             fault = jsonrpclib.Fault(-32603, 'Server error: {0}'
                                      .format(trace_string), config=config)
             _logger.exception("Server-side error: %s", fault)
@@ -416,9 +469,8 @@ class SimpleJSONRPCRequestHandler(xmlrpcserver.SimpleXMLRPCRequestHandler):
         self.send_header("Content-type", config.content_type)
         self.send_header("Content-length", str(len(response)))
         self.end_headers()
-        self.wfile.write(response)
-        self.wfile.flush()
-        self.connection.shutdown(1)
+        if response:
+            self.wfile.write(response)
 
 # ------------------------------------------------------------------------------
 
@@ -480,6 +532,59 @@ class SimpleJSONRPCServer(socketserver.TCPServer, SimpleJSONRPCDispatcher):
 # ------------------------------------------------------------------------------
 
 
+class PooledJSONRPCServer(SimpleJSONRPCServer, socketserver.ThreadingMixIn):
+    """
+    JSON-RPC server based on a thread pool
+    """
+    def __init__(self, addr, requestHandler=SimpleJSONRPCRequestHandler,
+                 logRequests=True, encoding=None, bind_and_activate=True,
+                 address_family=socket.AF_INET,
+                 config=jsonrpclib.config.DEFAULT, thread_pool=None):
+        """
+        Sets up the server and the dispatcher
+
+        :param addr: The server listening address
+        :param requestHandler: Custom request handler
+        :param logRequests: Flag to(de)activate requests logging
+        :param encoding: The dispatcher request encoding
+        :param bind_and_activate: If True, starts the server immediately
+        :param address_family: The server listening address family
+        :param config: A JSONRPClib Config instance
+        :param thread_pool: A ThreadPool object. The pool must be started.
+        """
+        # Normalize the thread pool
+        if thread_pool is None:
+            # Start a thread pool with  30 threads max, 0 thread min
+            thread_pool = jsonrpclib.threadpool.ThreadPool(
+                30, 0, logname="PooledJSONRPCServer")
+            thread_pool.start()
+
+        # Store the thread pool
+        self.__request_pool = thread_pool
+
+        # Prepare the server
+        SimpleJSONRPCServer.__init__(self, addr, requestHandler, logRequests,
+                                     encoding, bind_and_activate,
+                                     address_family, config)
+
+    def process_request(self, request, client_address):
+        """
+        Handle a client request: queue it in the thread pool
+        """
+        self.__request_pool.enqueue(self.process_request_thread,
+                                    request, client_address)
+
+    def server_close(self):
+        """
+        Clean up the server
+        """
+        SimpleJSONRPCServer.shutdown(self)
+        SimpleJSONRPCServer.server_close(self)
+        self.__request_pool.stop()
+
+# ------------------------------------------------------------------------------
+
+
 class CGIJSONRPCRequestHandler(SimpleJSONRPCDispatcher):
     """
     JSON-RPC CGI handler (and dispatcher)
